#include "ehs/io/audio/AudioDevice_PW.h"
#include "ehs/EHS.h"
#include "ehs/Log.h"
#include "ehs/system/Semaphore.h"

#include <spa/utils/result.h>
#include <spa/utils/dict.h>
#include <spa/buffer/buffer.h>
#include <spa/param/latency-utils.h>

namespace ehs
{
	Array<AudioDevice> AudioDevice::devices;

	void AudioDevice::RegistryEventGlobal(void *data, const UInt_32 id, UInt_32 permissions, const char *type, UInt_32 version, const spa_dict *props)
	{
		if (Str_8::Cmp(type, PW_TYPE_INTERFACE_Node))
		{
			const Str_8 mediaClass = spa_dict_lookup(props, PW_KEY_MEDIA_CLASS);
			if (mediaClass.Size() && (mediaClass == "Audio/Sink" || mediaClass == "Audio/Source"))
			{
				AudioDevice device;

				if (mediaClass == "Audio/Sink")
					device.type = AudioDeviceType::OUTPUT;
				else if (mediaClass == "Audio/Source")
					device.type = AudioDeviceType::INPUT;

				device.id = id;
				device.name = spa_dict_lookup(props, PW_KEY_NODE_NAME);

				EHS_LOG_INT(LogType::INFO, 1, "\nDevice Name: " + device.name + "\nId: " + Str_8::FromNum(device.id));

				devices.Push(device);
			}
		}
	}

	void AudioDevice::RegistryEventGlobalRemove(void *data, UInt_32 id)
	{
	}

	void AudioDevice::OnParamChanged(void *data, UInt_32 id, const spa_pod *param)
	{
		if (!param)
			return;

		if (id == SPA_PARAM_Buffers)
		{
			AudioDevice *device = (AudioDevice *)data;

			spa_pod_object *obj = (spa_pod_object *)param;
			spa_pod_prop *prop;

			SPA_POD_OBJECT_FOREACH(obj, prop)
			{
				if (prop->key == SPA_PARAM_BUFFERS_size)
				{
					spa_pod_get_int(&prop->value, (int32_t *)&device->maxFrames);
					device->maxFrames = device->maxFrames / device->byteDepth / device->channels;
				}
			}
		}
	}

	AudioDevice::~AudioDevice()
	{
		if (!IsStreaming())
			return;

		if (output)
		{
			pw_stream_disconnect(output);
			pw_stream_destroy(output);
			output = nullptr;
		}

		if (input)
		{
			pw_stream_disconnect(input);
			pw_stream_destroy(input);
			input = nullptr;
		}

		pw_core_disconnect(core);
		core = nullptr;

		pw_context_destroy(context);
		context = nullptr;

		pw_loop_destroy(loop);
		loop = nullptr;
	}

	AudioDevice::AudioDevice()
		: id(0), loop(nullptr), context(nullptr), core(nullptr), input(nullptr), output(nullptr), paramsHook{}
	{
	}

	AudioDevice::AudioDevice(AudioDevice&& device) noexcept
		: BaseAudioDevice(device), id(device.id), name((Str_8 &&)device.name), loop(device.loop),
		context(device.context), core(device.core), input(device.input), output(device.output), paramsHook(device.paramsHook)
	{
		device.id = 0;
		device.loop = nullptr;
		device.context = nullptr;
		device.core = nullptr;
		device.input = nullptr;
		device.output = nullptr;
		device.paramsHook = {};
	}

	AudioDevice::AudioDevice(const AudioDevice& device)
		: BaseAudioDevice(device), id(device.id), name(device.name), loop(nullptr), context(nullptr), core(nullptr),
		input(nullptr), output(nullptr), paramsHook{}
	{
	}

	AudioDevice& AudioDevice::operator=(AudioDevice&& device) noexcept
	{
		if (this == &device)
			return *this;

		BaseAudioDevice::operator=(device);

		id = device.id;
		name = (Str_8 &&)device.name;
		loop = device.loop;
		context = device.context;
		core = device.core;
		input = device.input;
		output = device.output;
		paramsHook = device.paramsHook;

		device.id = 0;
		device.loop = nullptr;
		device.context = nullptr;
		device.core = nullptr;
		device.input = nullptr;
		device.output = nullptr;
		device.paramsHook = {};

		return *this;
	}

	AudioDevice& AudioDevice::operator=(const AudioDevice& device)
	{
		if (this == &device)
			return *this;

		BaseAudioDevice::operator=(device);

		id = device.id;
		name = device.name;
		loop = nullptr;
		context = nullptr;
		core = nullptr;
		input = nullptr;
		output = nullptr;
		paramsHook = {};

		return *this;
	}

	void AudioDevice::OpenStream()
	{
		if (!IsValid() || IsStreaming())
			return;

		loop = pw_loop_new(nullptr);
		context = pw_context_new(loop, nullptr, 0);
		core = pw_context_connect(context, nullptr, 0);

		spa_audio_info_raw info = {
			.format = SPA_AUDIO_FORMAT_F32,
			.rate = GetSampleRate(),
			.channels = GetChannels(),
			.position = {
				SPA_AUDIO_CHANNEL_FL,
				SPA_AUDIO_CHANNEL_FR,
				SPA_AUDIO_CHANNEL_FC,
				SPA_AUDIO_CHANNEL_LFE,
				SPA_AUDIO_CHANNEL_SL,
				SPA_AUDIO_CHANNEL_SR,
				SPA_AUDIO_CHANNEL_RL,
				SPA_AUDIO_CHANNEL_RR
			}
		};

		pw_properties *props = pw_properties_new(
			PW_KEY_MEDIA_TYPE, "Audio",
			PW_KEY_MEDIA_CATEGORY, GetCategory(),
			PW_KEY_MEDIA_ROLE, "Game",
			nullptr
		);

		spa_pod_builder b = {};
		Byte buffer[1024];

		spa_pod_builder_init(&b, buffer, sizeof(buffer));

		const spa_pod *pod[] = {spa_format_audio_raw_build(&b, SPA_PARAM_EnumFormat, &info)};

		if (GetType() == AudioDeviceType::INPUT || GetType() == AudioDeviceType::ALL)
		{
			input = pw_stream_new(core, "", props);
			if (!input)
			{
				EHS_LOG_INT(LogType::ERR, 1, "Failed to create input stream for audio device.");
				return;
			}

			SInt_32 err = pw_stream_connect(
				input,
				PW_DIRECTION_INPUT,
				id,
				(pw_stream_flags)(PW_STREAM_FLAG_AUTOCONNECT | PW_STREAM_FLAG_MAP_BUFFERS),
				pod,
				1
			);
			if (err)
			{
				EHS_LOG_INT(LogType::ERR, 2, "Failed to connect input stream for audio device.");
			}
		}

		if (GetType() == AudioDeviceType::OUTPUT || GetType() == AudioDeviceType::ALL)
		{
			output = pw_stream_new(core, "", props);
			if (!output)
			{
				EHS_LOG_INT(LogType::ERR, 1, "Failed to create output stream for audio device.");
				return;
			}

			SInt_32 err = pw_stream_connect(
				output,
				PW_DIRECTION_OUTPUT,
				id,
				(pw_stream_flags)(PW_STREAM_FLAG_AUTOCONNECT | PW_STREAM_FLAG_MAP_BUFFERS),
				pod,
				1
			);
			if (err)
			{
				EHS_LOG_INT(LogType::ERR, 2, "Failed to connect output stream for audio device.");
			}
		}

		static constexpr pw_stream_events streamEvents = {
			.version = PW_VERSION_STREAM_EVENTS,
			.param_changed = OnParamChanged
		};

		pw_stream_add_listener(output, &paramsHook, &streamEvents, this);

		while (pw_stream_get_state(output, nullptr) == PW_STREAM_STATE_CONNECTING)
		{
			SInt_32 err = pw_loop_iterate(loop, 0);
			if (err < 0)
			{
				EHS_LOG_INT(LogType::ERR, 3, "Failed to update audio with error #" + Str_8::FromNum(err) + ".");
				return;
			}
		}

		EHS_LOG_SUCCESS();
	}

	void AudioDevice::CloseStream()
	{
		if (!IsStreaming())
			return;

		if (output)
		{
			pw_stream_disconnect(output);
			pw_stream_destroy(output);
			output = nullptr;
		}

		if (input)
		{
			pw_stream_disconnect(input);
			pw_stream_destroy(input);
			input = nullptr;
		}

		pw_core_disconnect(core);
		core = nullptr;

		pw_context_destroy(context);
		context = nullptr;

		pw_loop_destroy(loop);
		loop = nullptr;
	}

	UInt_64 AudioDevice::SendStream(const void *data, const UInt_64 size)
	{
		if (GetType() != AudioDeviceType::OUTPUT && GetType() != AudioDeviceType::ALL)
		{
			EHS_LOG_INT(LogType::ERR, 0, "You can only send an audio stream on an output audio device.");
			return 0;
		}

		if (!data || !size)
			return 0;

		SInt_32 err = pw_loop_iterate(loop, 0);
		if (err < 0)
		{
			EHS_LOG_INT(LogType::ERR, 1, "Failed to update audio with error #" + Str_8::FromNum(err) + ".");
			return 0;
		}

		if (pw_stream_get_state(output, nullptr) != PW_STREAM_STATE_STREAMING)
			return 0;

		pw_buffer *buf = pw_stream_dequeue_buffer(output);
		if (!buf)
			return 0;

		UInt_64 out = 0;

		spa_data &result = buf->buffer->datas[0];

		if (size > result.maxsize)
			out = result.maxsize;
		else
			out = size;

		Util::Copy(result.data, data, out);

		pw_stream_queue_buffer(output, buf);

		EHS_LOG_SUCCESS();

		return out;
	}

	UInt_64 AudioDevice::ReceiveStream(void *data, const UInt_64 size)
	{
		if (GetType() != AudioDeviceType::INPUT && GetType() != AudioDeviceType::ALL)
		{
			EHS_LOG_INT(LogType::ERR, 0, "You can only send an audio stream on an input audio device.");
			return 0;
		}

		pw_loop_iterate(loop, 0);

		if (pw_stream_get_state(input, nullptr) != PW_STREAM_STATE_STREAMING)
			return 0;

		pw_buffer *buf = pw_stream_dequeue_buffer(input);
		if (!buf)
			return 0;

		spa_data &result = buf->buffer->datas[0];

		UInt_64 in;
		if (result.chunk->size <= size)
			in = result.chunk->size;
		else
			in = size;

		Util::Copy(data, result.data, in);

		pw_stream_queue_buffer(input, buf);

		EHS_LOG_SUCCESS();

		return in;
	}

	bool AudioDevice::IsStreaming() const
	{
		return loop && context && core && (input || output);
	}

	bool AudioDevice::IsValid() const
	{
		return id;
	}

	AudioDevice AudioDevice::GetDefault(const AudioDeviceType type)
	{
		pw_init(nullptr, nullptr);

		AudioDevice result;
		result.type = type;
		result.id = PW_ID_ANY;

		return result;
	}

	Array<AudioDevice> AudioDevice::Get(const AudioDeviceType type, const AudioDeviceState state)
	{
		pw_init(nullptr, nullptr);

		pw_loop *loop = pw_loop_new(nullptr);
		pw_context *context = pw_context_new(loop, nullptr, 0);
		pw_core *core = pw_context_connect(context, nullptr, 0);
		pw_registry *registry = pw_core_get_registry(core, PW_VERSION_REGISTRY, 0);
		spa_hook listener = {};

		static constexpr pw_registry_events events = {
			.version = PW_VERSION_REGISTRY_EVENTS,
			.global = RegistryEventGlobal,
			.global_remove = RegistryEventGlobalRemove
		};

		pw_registry_add_listener(registry, &listener, &events, registry);

		while (pw_loop_iterate(loop, 10));

		pw_core_disconnect(core);
		pw_context_destroy(context);
		pw_loop_destroy(loop);

		return devices;
	}

	Str_8 AudioDevice::GetCategory() const
	{
		switch (GetType())
		{
			case AudioDeviceType::INPUT:
				return "Capture";
			case AudioDeviceType::OUTPUT:
				return "Playback";
			case AudioDeviceType::ALL:
				return "Duplex";
			default:
				return "";
		}
	}
}
